import pandas as pd
import numpy as np
import torch
from scipy import sparse
import itertools
import pickle
from collections.abc import Iterable

class GeneToPhenotypeDataset:

	def _process_gene_table(self):

		#possible dtypes = [float_array,one_hot_cat,sparse_binary,float]
		self.aux_gene_info_table=pd.DataFrame([],index=self.gene_info_table.columns,columns=['dtype','idx_map','inv_idx_map','return_dim'])
		for col in self.gene_info_table.columns:
			has_iterables = self.gene_info_table[col].apply(lambda x:isinstance(x,Iterable))
			has_strings = self.gene_info_table[col].apply(lambda x:isinstance(x,str))
			has_floats = self.gene_info_table[col].apply(lambda x:isinstance(x,float))
			if (has_iterables.sum()>0) and (has_strings.sum()==0):
				assert has_iterables.sum()==self.gene_info_table.shape[0], "Mixed iterables and other data types in gene_info_table columns is not allowed."
				#check length, if all the same, then must be float array. Otherwise, assume sparse-binary array
				iterable_lengths = self.gene_info_table[col].apply(lambda x:len(x))
				if np.unique(iterable_lengths.values).shape[0]==1:
					assert self.gene_info_table[col].apply(lambda lst: all((isinstance(x, float) for x in lst))).sum()==self.gene_info_table.shape[0], "If all entries in data column are the same length, only float entries are allowed. Arrays of ints should be re-coded as categorical variables."
					self.aux_gene_info_table.loc[col,'dtype']='float_array'
					self.aux_gene_info_table.loc[col]['idx_map']='NA'
					self.aux_gene_info_table.loc[col]['inv_idx_map']='NA'
					self.aux_gene_info_table.loc[col,'return_dim']=np.unique(iterable_lengths.values)[0]
				else:
					all_unique=sorted(list(set().union(*self.gene_info_table[col])))
					idx_map=dict(zip(all_unique,list(range(len(all_unique)))))
					self.aux_gene_info_table.loc[col,'dtype']='sparse_binary'
					self.aux_gene_info_table.loc[col]['idx_map']=idx_map
					self.aux_gene_info_table.loc[col]['inv_idx_map']=dict(zip(idx_map.values(),idx_map.keys()))
					self.gene_info_table[col]=self.gene_info_table[col].apply(lambda x: [idx_map[y] for y in x])
					self.aux_gene_info_table.loc[col,'return_dim']=len(idx_map)

			else:
				assert (has_strings.sum()==self.gene_info_table.shape[0]) or (has_floats.sum()==self.gene_info_table.shape[0]),"Single entry columns in gene_info_table must be either all float or all string. Mixed types or integers are not allowed."
				if has_floats.sum()==self.gene_info_table.shape[0]:
					self.aux_gene_info_table.loc[col,'dtype']='float'
					self.aux_gene_info_table.loc[col,'idx_map']='NA'
					self.aux_gene_info_table.loc[col]['inv_idx_map']='NA'
					self.aux_gene_info_table.loc[col,'return_dim']=1
				else:
					self.aux_gene_info_table.loc[col,'dtype']='one_hot_cat'
					all_unique=sorted(list(self.gene_info_table[col].unique()))
					assert len(all_unique)>1, "Must have at least 2 categories for one-hot variable."
					idx_map=dict(zip(all_unique,list(range(len(all_unique)))))
					self.aux_gene_info_table.loc[col]['idx_map']=idx_map
					self.aux_gene_info_table.loc[col]['inv_idx_map']=dict(zip(idx_map.values(),idx_map.keys()))
					self.gene_info_table[col]=self.gene_info_table[col].apply(lambda x: idx_map[x])
					self.aux_gene_info_table.loc[col,'return_dim']=len(idx_map)-1
	

	def __init__(self,disease_labels,symptom_freq_pairs,disease_associated_genes,gene_info_table,ordinal_freq_map,training_data_fraction,validation_data_fraction,missing_label='NA'):
		"""Dataset that stores disease-symptom annotations as a sparse array that can be used for disease embedding.
		
		Args:
		    disease_labels (Array of Strings): Strings that make up the index of the disease dataset
		    symptom_freq_pairs (Array of iterables): Each list contains the symptoms-frequency pairs annotated to each disease. Symptoms-frequencies are stored as a tuple.
		    ordinal_freq_map (Dictionary): Dictionary providing ordinal rank of each frequency used to describe the symptoms. 
		    missing_label (str, optional): String used to denote missing frequency information.
		"""
		self.disease_table=pd.DataFrame(columns=['SympFreqs','MappedGenes'],index=disease_labels)
		orig_symp_values = pd.Series(symptom_freq_pairs,index=disease_labels)
		all_unique_symptoms=sorted(list(set().union(*orig_symp_values.apply(lambda x:[y[0] for y in x]).values)))
		self.symptom_map=dict(zip(all_unique_symptoms,range(len(all_unique_symptoms))))
		self.inverse_symptom_map=dict(zip(self.symptom_map.values(),self.symptom_map.keys()))
		self.num_symptoms=len(self.symptom_map)
		
		self.ordinal_freq_map=ordinal_freq_map
		assert -1 not in ordinal_freq_map.values(),"-1 is not allowed as ordinal value for symptom frequency. This represents missing data and is reserved for this purpose."
		self.ordinal_freq_map[missing_label]=-1
		self.num_ordinal_freqs = len(self.ordinal_freq_map)-1
		self.missing_label=missing_label

		self.gene_info_table=gene_info_table

		new_symptoms = orig_symp_values.apply(lambda x: [(self.symptom_map[y[0]],self.ordinal_freq_map[y[1]]) for y in x])
		self.disease_table['SympFreqs']=new_symptoms
		orig_gene_values=pd.Series(disease_associated_genes,index=disease_labels)
		self.gene_map=dict(zip(self.gene_info_table.index,range(self.gene_info_table.index.shape[0])))
		self.inverse_gene_map=dict(zip(self.gene_map.values(),self.gene_map.keys()))
		self.num_total_genes = len(self.gene_map)
		self.disease_table['MappedGenes']=orig_gene_values.apply(lambda x: self.gene_map[x])


		self._process_gene_table()
		self.SetNewTrainingState(training_data_fraction,validation_data_fraction)



	def SetNewTrainingState(self, training_fraction, validation_fraction):

		assert (training_fraction+validation_fraction)<=1.0,"Training and validation fraction added together cannot exceed 1.0"
		if (training_fraction+validation_fraction)==1.0:
		    print("Warning: Setting a model training state without allowing for a test fraction. There will be test fraction available for independent replication.")

		num_training_diseases=int(np.floor(training_fraction*self.disease_table.shape[0]))
		num_validation_diseases=int(np.ceil(validation_fraction*self.disease_table.shape[0]))
		num_testing_diseases=int(self.disease_table.shape[0]-num_training_diseases-num_validation_diseases)

		self.training_index=pd.Index(np.random.choice(self.disease_table.index,size=num_training_diseases,replace=False))
		if (num_training_diseases+num_validation_diseases)==self.disease_table.index.shape[0]:
		    self.validation_index=self.disease_table.index.difference(self.training_index)
		else:
		    self.validation_index=pd.Index(np.random.choice(self.disease_table.index.difference(self.training_index),size=num_validation_diseases,replace=False))

		self.testing_index=self.disease_table.index.difference(np.union1d(self.training_index,self.validation_index))

		self.training_index=self.training_index.values
		self.validation_index=self.validation_index.values
		self.testing_index=self.testing_index.values

	def ShuffleTrainingValidation(self):
		combined_index=np.concatenate([self.training_index,self.validation_index])
		np.random.shuffle(combined_index)
		self.training_index=combined_index[0:self.training_index.shape[0]]
		self.validation_index=combined_index[self.training_index.shape[0]:]

	def SaveTrainingState(self,fName):
		if fName[-4:]!='.pth':
			fName+='.pth'
		currentState = dict()
		currentState['training_index']=self.training_index
		currentState['validation_index']=self.validation_index
		currentState['testing_index']=self.testing_index

		with open(fName, 'wb') as f:
			pickle.dump(currentState,f)

	def LoadTrainingState(self,fName):
		if fName[-4:]!='.pth':
			fName+='.pth'

		with open(fName, 'rb') as f:
			currentState = pickle.load(f)

		self.training_index=currentState['training_index']
		self.validation_index=currentState['validation_index']
		self.testing_index=currentState['testing_index']

		assert np.setdiff1d(np.union1d(np.union1d(self.training_index,self.validation_index),self.testing_index),self.disease_table.index.values).shape[0]==0,"Index of stored and loaded state do not match."




	def _torchWrapper(self,x):
		"""
		Note, all torch floating point tensors are converted to 32-bits to
		ensure GPU compatibility.
		"""

		if x.dtype==np.float32:
		    if sparse.issparse(x):
		        return torch.tensor(x.toarray(),dtype = torch.float32)
		    else:
		        return torch.tensor(x,dtype = torch.float32)

		elif x.dtype==np.float64:
		    if sparse.issparse(x):
		        return torch.tensor(x.toarray(),dtype = torch.float32)
		    else:
		        return torch.tensor(x,dtype = torch.float32)
		else:
		    if sparse.issparse(x):
		        return torch.tensor(x.toarray(),dtype = torch.long)
		    else:
		        return torch.tensor(x,dtype = torch.long)


	def _build_disease_coo_matrices(self,index):

		symp_x_inds=list(itertools.chain.from_iterable([[i]*len(x) for i,x in enumerate(self.disease_table.loc[index]['SympFreqs'])]))
		symp_y_inds=list(itertools.chain.from_iterable(self.disease_table.loc[index]['SympFreqs'].apply(lambda x: [y[0] for y in x])))
		values_freqs= list(itertools.chain.from_iterable(self.disease_table.loc[index]['SympFreqs'].apply(lambda x: [y[1] for y in x])))
		values_symps= np.ones((len(symp_x_inds)))

		return sparse.coo_matrix((values_symps,(symp_x_inds,symp_y_inds)),shape=(len(index),len(self.symptom_map)),dtype=np.float32),sparse.coo_matrix((values_freqs,(symp_x_inds,symp_y_inds)),shape=(len(index),len(self.symptom_map)),dtype=np.float32)

	def _build_gene_table_matrices(self, gene_list):
		return_data=[]
		for dcol in self.aux_gene_info_table.index:
			if self.aux_gene_info_table.loc[dcol]['dtype']=='float':
				return_data+=[self.gene_info_table.loc[gene_list][dcol].values.reshape(-1,1)]
			elif self.aux_gene_info_table.loc[dcol]['dtype']=='one_hot_cat':
				x_inds=list(itertools.chain.from_iterable([[i] for i,x in enumerate(self.gene_info_table.loc[gene_list][dcol]) if x!=0]))
				y_inds=self.gene_info_table.loc[gene_list][dcol].values
				y_inds=list(y_inds[y_inds!=0]-1)
				values=np.ones((len(x_inds)))
				return_data+=[sparse.coo_matrix((values,(x_inds,y_inds)),shape=(len(gene_list),self.aux_gene_info_table.loc[dcol]['return_dim']),dtype=np.float32)]
			elif self.aux_gene_info_table.loc[dcol]['dtype']=='float_array':
				return_data+=[np.vstack(self.gene_info_table.loc[gene_list][dcol].values)]
			else:
				x_inds=list(itertools.chain.from_iterable([[i]*len(x) for i,x in enumerate(self.gene_info_table.loc[gene_list][dcol])]))
				y_inds=list(itertools.chain.from_iterable(self.gene_info_table.loc[gene_list][dcol]))
				values= np.ones((len(x_inds)))
				return_data+=[sparse.coo_matrix((values,(x_inds,y_inds)),shape=(len(gene_list),self.aux_gene_info_table.loc[dcol]['return_dim']),dtype=np.float32)]
		return tuple(return_data)

	def ReturnDataArrays(self,index):
		symp_arrays=self._build_disease_coo_matrices(index)
		symp_arrays=tuple(self._torchWrapper(x) for x in symp_arrays)
		gene_arrays=self.ReturnGeneDataArrays([self.inverse_gene_map[x] for x in self.disease_table.loc[index]['MappedGenes'].values])
		return symp_arrays+(gene_arrays,)

	def ReturnGeneDataArrays(self,gene_list):
		gene_arrays=self._build_gene_table_matrices(gene_list)
		gene_arrays=tuple(self._torchWrapper(x) for x in gene_arrays)
		return torch.cat(gene_arrays,axis=1)


	def DropSymptoms(self,symptom_list):

		oldSymptomToIntMap=self.symptom_map

		allSymptoms=set(oldSymptomToIntMap.keys())

		removedSymptoms=set(symptom_list)

		assert len(removedSymptoms.difference(allSymptoms))==0, "Symptoms: {0:s} not in set of possible symptoms.".format(','.join(list(removedSymptoms.difference(allSymptoms))))

		keptInts = set([oldSymptomToIntMap[x] for x in allSymptoms.difference(symptom_list)])


		for old_symp in removedSymptoms:
		    del oldSymptomToIntMap[old_symp]

		newSymptomToIntMap = {}
		oldToNewIntMap={}
		for i,key in enumerate(oldSymptomToIntMap):
			oldToNewIntMap[oldSymptomToIntMap[key]]=i
			newSymptomToIntMap[key] = i


		self.symptom_map=newSymptomToIntMap
		self.inverse_symptom_map=dict(zip(self.symptom_map.values(),self.symptom_map.keys()))
		self.disease_table['SympFreqs']=self.disease_table['SympFreqs'].apply(lambda x: [(oldToNewIntMap[y[0]],y[1]) for y in x if y[0] in keptInts])

	def DropDiseases(self, disease_list):
		disease_list=pd.Index(disease_list)
		assert len(disease_list.difference(self.disease_table.index))==0,"Subjects: {0:s} not in data table.".format(','.join(list(disease_list.difference(self.dataset.index))))
		self.disease_table.drop(index=disease_list,inplace=True)


	def FindAllDiseases_wSymptom(self,symptom):
		if symptom not in self.symptom_map.keys():
			raise KeyError("Symptom {0:s} not in possible set.".format(symptom))
		else:
			intVal = self.symptom_map[symptom]
			return self.disease_table.index[self.disease_table['SympFreqs'].apply(lambda x: intVal in set([y[0] for y in x]))]

	def FindAllDiseases_wGene(self,gene):
		if gene not in self.gene_map.keys():
			raise KeyError("Gene {0:s} not in possible set.".format(gene))
		else:
			intVal = self.gene_map[gene]
			return self.disease_table.index[self.disease_table['MappedGenes'].apply(lambda x: intVal==x)]

	def InferNullModel(self,pseudo_count=1.0):
		training_data_symps,training_data_freqs = self._build_disease_coo_matrices(self.training_index)
		base_symp_freqs=((pseudo_count/2.0)+ np.array(training_data_symps.sum(axis=0)).ravel())/(pseudo_count+training_data_symps.shape[0])
		missing_rates=((pseudo_count/2.0)+ np.array((training_data_symps==1).multiply(training_data_freqs==-1).sum(axis=0)).ravel())/(pseudo_count+np.array(training_data_symps.sum(axis=0)).ravel())

		full_freq_array=training_data_freqs.toarray()
		full_freq_array[training_data_symps.toarray()==0.0]=-1.0
		full_freq_array[full_freq_array==-1]=full_freq_array.max()+1
		unrolled_freq_array=np.zeros((full_freq_array.shape[0],full_freq_array.shape[1],int(full_freq_array.max()+1)))
		for i in range(unrolled_freq_array.shape[0]):
			unrolled_freq_array[i,np.arange(unrolled_freq_array.shape[1]),np.array(full_freq_array[i],dtype=np.int32)]=1
		unrolled_freq_array=unrolled_freq_array[:,:,:-1]
		freq_counts=unrolled_freq_array.sum(axis=0)+(pseudo_count/unrolled_freq_array.shape[2])
		freq_stats=freq_counts/freq_counts.sum(axis=1).reshape(freq_counts.shape[0],-1)
		return {'BaselineSymptomRates':base_symp_freqs,"MissingRates":missing_rates,'FrequencyRates':freq_stats}
			
if __name__=='__main__':
	omim_table=pd.read_pickle('../../../Data/OMIM_HPO_Table/OMIMSingleGeneInNetworkHPOUnrolled.pth')
	pp_embed=pd.read_csv('../../../Data/PrimeKG/NetworkEmbeddings/PrimeKG_protein_protein_Dim8.txt',sep=' ',header=None,skiprows=[0])
	pp_embed.set_index(0,inplace=True)
	gene_table=pd.read_pickle('../../../Data/PrimeKG/PrimeKG_ProteinFeatures.pth')
	gene_table.loc[gene_table.index,'net_embed']=pd.Series([x for x in pp_embed.loc[gene_table.index].values],index=gene_table.index)


	self=GeneToPhenotypeDataset(omim_table.index,omim_table.HPO_wFreq,omim_table['GENE/INHERIT'].apply(lambda x:x[0]),gene_table,{'VR':0,'OC':1,'F':2,'VF':3,'O':4},0.6,0.15,missing_label='NA')



import torch
from torch import nn
import pyro
import pyro.distributions as dist
import copy
from pyro.infer import Trace_ELBO

from .networks import FC_MeanScaleEncoder,FC_Decoder,Linear_Decoder

class DiseaseCondVAE(nn.Module):

	def __init__(self,
		num_symptoms,
		n_freq_classes,
		n_latent_dim,
		baseline_g2p_model,
		isLinear=True,
		encoder_hyperparameters={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.1, 'use_batch_norm':True},
		decoder_hyperparameters={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.1, 'use_batch_norm':True},
		missing_freq_prior_mean=0.0,
		missing_freq_prior_scale=3.0,
		cut_points = [0.04,0.3,0.8,0.99]
		):    

		super(DiseaseCondVAE,self).__init__()
		self.num_symptoms=num_symptoms
		self.n_latent_dim=n_latent_dim
		self.n_freq_classes=n_freq_classes
		self.baseline_g2p_model=baseline_g2p_model


		#encoder receives the following information: annotated symptoms, whether frequency is present or missing, and the assigned frequency. This information can be encoded using self.num_dx (annotation present)+self.num_dx (frequency present vs missing)+ self.num_dx*(self.n_freq_classes) (using one-hot endoding)
		self.encoder=FC_MeanScaleEncoder((self.n_freq_classes+2)*self.num_symptoms+self.baseline_g2p_model.g2p_dataset.aux_gene_info_table['return_dim'].sum(),n_latent_dim,**encoder_hyperparameters)
		self.prior_model=FC_MeanScaleEncoder(3*self.num_symptoms+self.baseline_g2p_model.g2p_dataset.aux_gene_info_table['return_dim'].sum(),n_latent_dim,**encoder_hyperparameters)

		self.isLinear=isLinear

		if self.isLinear:
			self.symptom_annotation_decoder=Linear_Decoder(n_latent_dim,self.num_symptoms)
			self.symptom_frequency_decoder=Linear_Decoder(n_latent_dim,self.num_symptoms)
		else:
			self.symptom_annotation_decoder=FC_Decoder(self.n_latent_dim,self.num_symptoms,**decoder_hyperparameters)
			self.symptom_frequency_decoder=FC_Decoder(self.n_latent_dim,self.num_symptoms,**decoder_hyperparameters)

		self.missing_freq_disease_decoder=nn.Linear(self.n_latent_dim,1,bias=True)

		self.missing_freq_prior_scale=torch.tensor(missing_freq_prior_scale,dtype=torch.float32)
		self.missing_freq_prior_mean=torch.tensor(missing_freq_prior_mean,dtype=torch.float32)

		self.cut_points=torch.logit(torch.tensor(cut_points,dtype=torch.float32))

		assert (len(self.cut_points)+1)==self.n_freq_classes, "Number of cut points+1 must equal number of frequency classes"

		self.missing_freq_intercepts=torch.zeros(self.num_symptoms,dtype=torch.float32)

		self.missing_freq_intercepts_post_mean=torch.zeros(self.num_symptoms,dtype=torch.float32)
		self.missing_freq_intercepts_post_log_scale=torch.zeros(self.num_symptoms,dtype=torch.float32)

		
		self.compute_device='cpu'
		self.eval()

	def model(self,annot_data=None,freq_data=None,gene_data=None,num_samples=None,minibatch_scale=1.0, annealing_factor=1.0):
		if annot_data is not None:
			num_samples=annot_data.shape[0]
		else:
			if num_samples is None:
				num_samples=1000
				print('Warning: no arguments were given to DiseaseVAE.model. This should only be done during debugging.')
		if freq_data is not None:
			assert (freq_data.shape[0]==num_samples)&(freq_data.shape[1]==self.num_symptoms), "Frequency annotation data dimensions do not match number of samples/symptoms"

			freq_missing=torch.zeros(freq_data.shape,dtype=torch.float32,device=self.compute_device)
			freq_missing[(freq_data==-1.0)]=1.0

			freq_missing_mask = torch.zeros(freq_data.shape,dtype=torch.bool,device=self.compute_device)
			freq_obs_mask = torch.zeros(freq_data.shape,dtype=torch.bool,device=self.compute_device)
			freq_missing_mask[annot_data==1]=1.0
			freq_obs_mask[(annot_data==1)&(freq_data!=-1)]=1.0

			transformed_freq_data=freq_data.detach().clone().long()
			transformed_freq_data[(annot_data==1.0)&(freq_data==-1.0)]=0.0
		else:
			freq_missing=None
			freq_missing_mask=None
			freq_obs_mask=None
			transformed_freq_data=None



		pyro.module("symptom_annotation_decoder", self.symptom_annotation_decoder)
		pyro.module("symptom_frequency_decoder", self.symptom_frequency_decoder)
		pyro.module("missing_freq_disease_decoder", self.missing_freq_disease_decoder)
		pyro.module("prior_model",self.prior_model)

		with torch.no_grad():
			self.missing_freq_prior_scale=self.missing_freq_prior_scale.detach().to(self.compute_device)
			self.missing_freq_prior_mean=self.missing_freq_prior_mean.detach().to(self.compute_device)
			self.cut_points=self.cut_points.detach().to(self.compute_device)

		self.missing_freq_intercepts=pyro.sample("missing_freq_intercepts", dist.Normal(torch.zeros(self.num_symptoms,dtype=torch.float32,device=self.compute_device)+self.missing_freq_prior_mean, torch.ones(self.num_symptoms,dtype=torch.float32,device=self.compute_device)*self.missing_freq_prior_scale).to_event(1))


		with pyro.poutine.scale(None,minibatch_scale):

			with pyro.plate("data",size=num_samples,dim=-2):
				with torch.no_grad():
					gene_pred= self.baseline_g2p_model.basic_net.forward(gene_data)
					gene_pred=torch.cat(gene_pred,axis=1)
				z_loc,z_scale = self.prior_model(torch.cat((gene_pred,gene_data),axis=1))
				with pyro.poutine.scale(None, annealing_factor):
					latent_variables=pyro.sample("latent_variables",dist.Normal(z_loc,z_scale))


				annotation_pred = self.symptom_annotation_decoder.forward(latent_variables)
				annotation_outcomes = pyro.sample("annotation_outcomes",dist.Bernoulli(logits=annotation_pred),obs=annot_data)


				if freq_missing_mask is None:
					freq_missing_mask = torch.zeros(annotation_outcomes.shape,dtype=torch.bool,device=self.compute_device)
					freq_missing_mask[annotation_outcomes==1]=1

				with pyro.plate("freq_annots",size=self.num_symptoms,dim=-1):

					with pyro.poutine.mask(mask=freq_missing_mask):
						missing_pred = self.missing_freq_disease_decoder.forward(latent_variables)+self.missing_freq_intercepts
						missing_outcomes = pyro.sample("missing_outcomes",dist.Bernoulli(logits=missing_pred),obs=freq_missing)

					if freq_obs_mask is None:
						freq_obs_mask = torch.zeros(annotation_outcomes.shape,dtype=torch.bool,device=self.compute_device)
						freq_obs_mask[(annotation_outcomes==1)&(missing_outcomes!=-1)]=1

					with pyro.poutine.mask(mask=freq_obs_mask):
						frequency_predictions=self.symptom_frequency_decoder.forward(latent_variables)
						frequency_outcomes=pyro.sample("frequency_outcomes",dist.OrderedLogistic(frequency_predictions,self.cut_points),obs=transformed_freq_data)


		freq_data_output=torch.zeros(annotation_outcomes.shape,dtype=torch.long,device=self.compute_device)
		freq_data_output[(annotation_outcomes==1)]=frequency_outcomes[(annotation_outcomes==1)]
		freq_data_output[(annotation_outcomes==1)&(missing_outcomes==1)]=-1.0

		return annotation_outcomes,freq_data_output

	def guide(self,annot_data=None,freq_data=None,gene_data=None,num_samples=None,minibatch_scale=1.0, annealing_factor=1.0):
		if annot_data is not None:
			num_samples=annot_data.shape[0]
		else:
			if num_samples is None:
				num_samples=1000
				print('Warning: no arguments were given to ClinicalDxVAE.model. This should only be done during debugging.')


		if freq_data is not None:

			assert (freq_data.shape[0]==num_samples)& (freq_data.shape[1]==self.num_symptoms), "Frequency annotation data dimensions do not match number of samples/symptoms"
			freq_missing=torch.zeros(freq_data.shape,dtype=torch.float32,device=self.compute_device)
			freq_missing[(freq_data==-1.0)]=1.0

			freq_missing_mask = torch.zeros(freq_data.shape,dtype=torch.bool,device=self.compute_device)
			freq_obs_mask = torch.zeros(freq_data.shape,dtype=torch.bool,device=self.compute_device)
			freq_missing_mask[annot_data==1]=1.0
			freq_obs_mask[(annot_data==1)&(freq_data!=-1)]=1.0
		else:
			freq_missing_mask=torch.ones(annot_data.shape,dtype=torch.bool,device=self.compute_device)
			freq_obs_mask=torch.ones(annot_data.shape,dtype=torch.bool,device=self.compute_device)


		pyro.module("encoder", self.encoder)

		self.missing_freq_intercepts_post_mean = pyro.param('missing_freq_intercepts_post_mean', torch.zeros(self.num_symptoms,dtype=torch.float32,device=self.compute_device))
		self.missing_freq_intercepts_post_log_scale = pyro.param('missing_freq_intercepts_post_log_scale', torch.zeros(self.num_symptoms,dtype=torch.float32,device=self.compute_device))

		self.missing_freq_intercepts=pyro.sample("missing_freq_intercepts",dist.Normal(self.missing_freq_intercepts_post_mean,torch.exp(self.missing_freq_intercepts_post_log_scale)).to_event(1))

		with pyro.poutine.scale(None,minibatch_scale):
			#unroll the data
			missing_freqs=torch.zeros(annot_data.shape,dtype=torch.float32,device=self.compute_device)
			missing_freqs[freq_data==-1]=1.0

			#assign unnannotated and missing data to an unobserved class
			transformed_freq_data=freq_data.detach().clone().long()

			transformed_freq_data[freq_data==-1]=len(self.cut_points)+1
			transformed_freq_data[(annot_data==0)]=len(self.cut_points)+1

			unrolled_freq_classes = torch.nn.functional.one_hot(transformed_freq_data)

			input_data=torch.cat([annot_data,missing_freqs,unrolled_freq_classes[:,:,:-1].flatten(start_dim=1),gene_data],axis=1)
			z_mean,z_std = self.encoder.forward(input_data)
			with pyro.poutine.scale(None, annealing_factor):
				with pyro.plate("data",size=num_samples,dim=-2):
					pyro.sample("latent_variables", dist.Normal(z_mean, z_std))

	def posterior_latent_state(self,annot_data,freq_data,gene_data):
		if self.training:
			self.eval()
			in_training=True
		else:
			in_training=False

		missing_freqs=torch.zeros(annot_data.shape,dtype=torch.float32,device=self.compute_device)
		missing_freqs[freq_data==-1]=1.0

		#assign unnannotated and missing data to an unobserved class
		transformed_freq_data=freq_data.detach().clone().long()

		transformed_freq_data[freq_data==-1]=len(self.cut_points)+1
		transformed_freq_data[(annot_data==0)]=len(self.cut_points)+1

		unrolled_freq_classes = torch.nn.functional.one_hot(transformed_freq_data)

		input_data=torch.cat([annot_data,missing_freqs,unrolled_freq_classes[:,:,:-1].flatten(start_dim=1),gene_data],axis=1)


		p_mean,p_std=self.encoder.forward(input_data)
		if in_training:
			self.train()

		return p_mean.detach(),p_std.detach()

	def package_state(self):
		"""
		Packages the model state dict as a dictionary.

		Returns
		-------
		dict
		    Model state dict.

		"""
		packaged_model_state={}
		packaged_model_state['model_state'] = copy.deepcopy(self.state_dict(keep_vars=True))
		packaged_model_state['baseline_g2p_model'] = copy.deepcopy(self.baseline_g2p_model.basic_net.state_dict(keep_vars=True))
		packaged_model_state['variational_post_params']={}
		packaged_model_state['variational_post_params']['missing_freq_intercepts_post_mean']=self.missing_freq_intercepts_post_mean.detach()
		packaged_model_state['variational_post_params']['missing_freq_intercepts_post_log_scale']=self.missing_freq_intercepts_post_log_scale.detach()
		
		return packaged_model_state

	def load_state(self,prior_model_state):
		"""
		Loads model state from dictionary

		Parameters
		----------
		prior_model_state : dict
		    Dictionary of model state produced by PackageCurrentState

		Returns
		-------
		None

		"""
		self.load_state_dict(prior_model_state['model_state'],strict=True)
		self.baseline_g2p_model.basic_net.load_state_dict(prior_model_state['baseline_g2p_model'],strict=True)
		self.missing_freq_intercepts_post_mean=prior_model_state['variational_post_params']['missing_freq_intercepts_post_mean']
		self.missing_freq_intercepts_post_log_scale=prior_model_state['variational_post_params']['missing_freq_intercepts_post_log_scale']

	def switch_device(self, new_compute_device):
		"""Sw

		Parameters
		----------
		compute_device : str
		    String name for compute device. Ideally expressed as 'type:number'
		"""
		self.compute_device=new_compute_device
		self.to(self.compute_device)
		self.missing_freq_intercepts_post_mean.to(self.compute_device)
		self.missing_freq_intercepts_post_log_scale.to(self.compute_device)
		self.baseline_g2p_model.basic_net.to(self.compute_device)

	def per_datum_ELBO(self,annot_data,freq_data,gene_data,num_particles=10,prior_only=False):
		""" Computes the evidence lower bound (ELBO) for each observation in the dataset.

		Parameters
		----------
		obs_data : torch.tensor
		    Binary array of observed data.
		num_particles : int
		    Number of particles (samples) used to approximate the ELBOs

		Returns
		-------
		torch.tensor
		    Per-datum ELBOs

		"""
		if self.training:
		    self.eval()
		    in_training=True
		else:
		    in_training=False

		elboFunc = Trace_ELBO(num_particles=num_particles)
		elboVec = torch.zeros(annot_data.shape[0],dtype=torch.float32,device=self.compute_device)

		for model_trace, guide_trace in elboFunc._get_traces(self.model, self.guide,(annot_data,freq_data,gene_data,),{}):
		    elboVec+=model_trace.nodes['annotation_outcomes']['log_prob'].detach().sum(axis=1)/num_particles
		    elboVec+=model_trace.nodes['missing_outcomes']['log_prob'].detach().sum(axis=1)/num_particles
		    elboVec+=model_trace.nodes['frequency_outcomes']['log_prob'].detach().sum(axis=1)/num_particles
		    elboVec+=model_trace.nodes['latent_variables']['log_prob'].detach().sum(axis=1)/num_particles
		    elboVec-=guide_trace.nodes['latent_variables']['log_prob'].detach().sum(axis=1)/num_particles

		if in_training:
		    self.train()
		return elboVec.reshape(elboVec.shape[0],1)



import numpy as np
import torch
import pyro
from pyro.infer import SVI
from pyro.infer import Trace_ELBO
from torch.utils import data
from pyro.optim import AdamW
from scipy.stats import linregress
from pyro import poutine
from collections import deque

from .data_wrapper import TorchDataWrapper

def rel_diff(curr,prev):
    return abs((curr - prev) / prev)

def collate_it(x):
    return x[0]


class AnnealingScheduler:
    def _cosine_anneal(self,currentStepFrac):
        cos_vals = np.cos(np.pi * currentStepFrac) + 1.0
        return self.finalTemp + ((self.initTemp - self.finalTemp) / 2.0) * cos_vals

    def _linear_anneal(self,currentStepFrac):
        return (self.finalTemp - self.initTemp) * currentStepFrac + self.initTemp

    def __init__(self,initTemp,finalTemp,totalNumSteps,scheduleType):
        """ Scheduler for KL-Anneling.

        Parameters
        ----------
        initTemp : float
            Initial temperature.
        finalTemp : float
            Final temperature.
        totalNumSteps : int
            Number of total steps for annealing
        scheduleType : str
            Type of scheduler. Must be 'linear' or 'cosine'.

        Returns
        -------
        None

        """

        assert initTemp>=0.0 and finalTemp >=0.0,"Annealing temperatures must be greater than 0.0"
        self.initTemp=initTemp
        self.finalTemp=finalTemp
        self.totalNumSteps=totalNumSteps
        if scheduleType=='linear':
            self._anneal_func=self._linear_anneal
        else:
            self._anneal_func=self._cosine_anneal

    def currentTemp(self,currentStep):
        frac = currentStep/self.totalNumSteps
        if frac>1.0:
            return self.finalTemp
        else:
            return self._anneal_func(frac)


class AutoEncoderOptimizer:

    def _returnELBOSlopeError(self,elbo_deque):
        """ Takes that average of the deque containing ELBO values. Helps determine convergence in the face of noisy SVI by taking the slope of the ELBO values as a function of iteration number. Error is computed as abs(slope)/mean(ELBO)
        
        Parameters
        ----------
        elbo_deque : collections.deque
            Deque containing previous N ELBO values (depends on deque size).
        
        Returns
        -------
        float
            ELBO error 
        """
        return abs(linregress(range(len(elbo_deque)),elbo_deque)[0]/np.mean(elbo_deque))




    def __init__(self,vae_model,dataset_training_state,optimizationParameters={'learningRate': 0.05,'maxEpochs': 500,'numParticles':10},computeConfiguration={'device':None,'numDataLoaders':0}, **kwargs):
        """

        This class implements a SGD optimizer, which uses mini-batches randomly sampled from a training dataset. To improve convergence rate and escape local minima, learning rate can be autommatically altered during inference. Note, given stochastic inference strategy coupled with possible re-starts, class tracks model perfomance and stores the best possible instance of the model obtained throughout the inference process.

                vae_model-->The vae model to optimize
                dataset_training_state-->DatasetTrainingState that holds clinical data and indices of the training and validation datasets.

                Other parameters are described in detail below and in the source code.
        Parameters
        ----------
        vae_model: nn.Module
            The vae model to optimize. 
        dataset_training_state: data.DatasetTrainingState
            DatasetTrainingState for some ClinicalDataset 
        optimizationParameters : dict
            Dictionary containing optimzation parameters. Default: {'learningRate': 0.05,'maxEpochs': 5000,'numParticles':10}
        computeConfiguration : dict
            Dictionary containing the compute configuration (device and number of dataloaders. Defaults to using cpu: {'device':None,'numDataLoaders':0}
        **kwargs : See source code.
            These arguments adjust the Adam weight decay parameter ('AdamW_Weight_Decay') and the KL-Annealing scheduler ('KLAnnealingParams'). See source code for details.

        Returns
        -------
        None

        """



        allKeywordArgs = list(kwargs.keys())

        self.vae_model=vae_model
        self.dataset_training_state=dataset_training_state

        #general optimization parameters
        self.maxEpochs=optimizationParameters['maxEpochs']
        self.learningRate = optimizationParameters['learningRate']
        self.numParticles = optimizationParameters['numParticles']

        #compute resources parameters
        self.device=computeConfiguration['device']
        if self.device is None:
            self.device='cpu'

        else:
            try:
                device_type,device_num=self.device.split(':')
            except ValueError:
                device_type=self.device
                device_num='0'
                self.device=device_type+':'+device_num
            assert device_type in ['cuda','mps','cpu']," Only 'cuda','mps', and 'cpu' supported"
            if device_type=='mps':
                assert torch.backends.mps.is_available()==True,"mps device not available"
            if device_type=='cuda':
                assert torch.cuda.is_available()==True,"cuda device is not available"

        self.num_dataloaders=computeConfiguration['numDataLoaders']


        if 'AdamW_Weight_Decay' in allKeywordArgs:
            self.AdamW_Weight_Decay=kwargs['AdamW_Weight_Decay']
        else:
            self.AdamW_Weight_Decay=1e-4

        if 'KLAnnealingParams' not in allKeywordArgs:
            self.KLAnnealingParams = {'initialTemp':1.0,'maxTemp':1.0,'fractionalDuration':1.0,'schedule': 'cosine'}
        else:
            self.KLAnnealingParams=kwargs['KLAnnealingParams']
            assert set(self.KLAnnealingParams.keys())==set(['initialTemp','maxTemp','fractionalDuration','schedule']),"KL Annealing Parameters must be dictionary with the following keys: 'initialTemp','maxTemp','fractionalDuration','schedule'"




    def BatchTrain(self,batch_size:int,errorTol:float = 1e-3,verbose=True,logFile=None,errorComputationWindow=None,early_stop_patience=5,monitor_validation=True):
        """

        Trains the VAE model using the mini-batches. This is the recommended method.

        Parameters
        ----------
        batch_size : int
            Number of samples in each mini-batch
        errorTol : float
            Error tolerance in validation data ELBO for convergence.
        verbose : bool
            Whether to print updates regarding fitting.
        logFile : str
            Path to logfile which can store progress
        errorComputationWindow : int
            Sliding window for computing error tolerance. Default appears to work well.

        Returns
        -------
        tuple
            (bestModelScore, vector of ELBO evaluations on training data, vector of ELBO evaluations on validation data, vector of all errors computed during fitting)

        """
        if errorComputationWindow is None:
            errorComputationWindow = max(int(0.01*self.maxEpochs),2)
        else:
            assert errorComputationWindow>0, "Expects errorComputationWindow to be integer > 0"

        error_window = deque(maxlen=errorComputationWindow)
        elbo_window = deque(maxlen=errorComputationWindow)
        errorVec = []
        trainLoss = []
        validationLoss = []

        if self.dataset_training_state.training_index is not None:
            numTotalTrainingSamples=self.dataset_training_state.training_index.shape[0]
            numTotalValidationSamples=self.dataset_training_state.validation_index.shape[0]
        else:
            raise ValueError("Training state for data.DatasetTrainingState not set prior to calling optimizer. Please set training set using data.DatasetTrainingState.SetNewTrainingState") 

        assert batch_size < numTotalTrainingSamples, "Batch size is greater than or equal to training data size."

        if self.device.split(':')[0]!='cpu':
            self.vae_model.switch_device(self.device)

            

        torchTrainingData=TorchDataWrapper(self.dataset_training_state,batch_size,index='Training')
        torchValidationData=TorchDataWrapper(self.dataset_training_state,batch_size,index='Validation')


        sviFunction = SVI(self.vae_model.model,self.vae_model.guide,AdamW({'weight_decay':self.AdamW_Weight_Decay,'lr':self.learningRate}),loss=Trace_ELBO(num_particles=self.numParticles))

        annealer=AnnealingScheduler(self.KLAnnealingParams['initialTemp'],self.KLAnnealingParams['maxTemp'],int(len(torchTrainingData)*(self.maxEpochs*self.KLAnnealingParams['fractionalDuration'])),self.KLAnnealingParams['schedule'])

        paramUpdateNum=0

        #initialize model states and scores
        bestModelState=self.vae_model.package_state()

        #note, batch_size handled at the level of _TorchDatasetWrapper, not loader itself.
        trainingDataLoader = data.DataLoader(torchTrainingData,num_workers=self.num_dataloaders,collate_fn=collate_it)
        validationDataLoader = data.DataLoader(torchValidationData,num_workers=self.num_dataloaders,collate_fn=collate_it)

        avg_epoch_train_loss = 0.0
        self.vae_model.eval()
        for i,data_batch in enumerate(trainingDataLoader):
            if self.device.split(':')[0]!='cpu':
                data_batch=tuple([x.to(self.device) for x in data_batch])

            avg_epoch_train_loss+=sviFunction.evaluate_loss(*data_batch,minibatch_scale = (numTotalTrainingSamples/data_batch[0].shape[0]),annealing_factor=annealer.currentTemp(paramUpdateNum+1))
        avg_epoch_train_loss=avg_epoch_train_loss/(i+1)
        prev_train_loss=avg_epoch_train_loss


        avg_epoch_val_loss = 0.0
        for i,data_batch in enumerate(validationDataLoader):
            if self.device.split(':')[0]!='cpu':
                data_batch=tuple([x.to(self.device) for x in data_batch])

            avg_epoch_val_loss+=sviFunction.evaluate_loss(*data_batch,minibatch_scale = (numTotalValidationSamples/data_batch[0].shape[0]),annealing_factor=self.KLAnnealingParams['maxTemp'])
        avg_epoch_val_loss = avg_epoch_val_loss/(i+1)
        bestModelScore = np.inf

        elbo_window.append(prev_train_loss)

        early_stop_count=0
        for epoch in range(self.maxEpochs):
            torchTrainingData.shuffle_index()
            avg_epoch_train_loss = 0.0
            self.vae_model.train()
            for i,data_batch in enumerate(trainingDataLoader):
                paramUpdateNum+=1
                if self.device.split(':')[0]!='cpu':
                    data_batch=tuple([x.to(self.device) for x in data_batch])
                avg_epoch_train_loss+=sviFunction.step(*data_batch,minibatch_scale = (numTotalTrainingSamples/data_batch[0].shape[0]),annealing_factor=annealer.currentTemp(paramUpdateNum))
            avg_epoch_train_loss=avg_epoch_train_loss/(i+1.0)

            avg_epoch_val_loss = 0.0
            self.vae_model.eval()
            for i,data_batch in enumerate(validationDataLoader):
                if self.device.split(':')[0]!='cpu':
                    data_batch=tuple([x.to(self.device) for x in data_batch])
                avg_epoch_val_loss+=sviFunction.evaluate_loss(*data_batch,minibatch_scale = (numTotalValidationSamples/data_batch[0].shape[0]),annealing_factor=self.KLAnnealingParams['maxTemp'])
            avg_epoch_val_loss=avg_epoch_val_loss/(i+1.0)

            if np.isnan(avg_epoch_val_loss) or np.isnan(avg_epoch_train_loss):
                print("Warning: NaN detected during inference. Model unlikely to be fully optimized!")
                break

            if monitor_validation:
                if avg_epoch_val_loss<bestModelScore:
                    bestModelState = self.vae_model.package_state()
                    bestModelScore = avg_epoch_val_loss
                    early_stop_count=0
                else:
                    early_stop_count+=1 
            else:
                if avg_epoch_train_loss<bestModelScore:
                    bestModelState = self.vae_model.package_state()
                    bestModelScore = avg_epoch_train_loss
                    early_stop_count=0
                else:
                    early_stop_count+=1 


            #track overall convergence
            trainLoss+=[avg_epoch_train_loss]
            validationLoss+=[avg_epoch_val_loss]



            error_window.append(rel_diff(avg_epoch_train_loss,prev_train_loss))
            elbo_window.append(avg_epoch_train_loss)

            avg_error = sum(error_window)/len(error_window)
            med_error = sorted(error_window)[int(0.5*len(error_window))]
            slope_error = self._returnELBOSlopeError(elbo_window)

            prev_train_loss=avg_epoch_train_loss
            errorVec+=[min([avg_error,med_error,slope_error])]

            if early_stop_count>early_stop_patience:
                if verbose:
                    print("Model diverging after %03d epochs; stopping optimization. Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Model diverging after %03d epochs; stopping optimization. Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                break


            if (avg_error < errorTol) or (med_error < errorTol) or (slope_error<errorTol):


                if verbose:
                    print("Optimization converged in %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Optimization converged in %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                break

            else:
                if verbose:
                    print("Completed %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Completed %03d epochs; Current Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))


        if epoch == (self.maxEpochs-1):
            f.write("Completed the maximum of %03d epochs; Final Loss (Train, Test): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))


        self.vae_model.load_state(bestModelState)
        #set eval mode
        if self.device.split(':')[0]!='cpu':
            self.vae_model.switch_device('cpu')
        self.vae_model.eval()
        return bestModelScore,trainLoss,validationLoss,errorVec





#!python


import setuptools
import re

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('VAE_G2P/vae_g2p.py').read(),
    re.M).group(1)



with open("README.md", "r") as fh:
    long_description = fh.read()



setuptools.setup(
    name="VAE_G2P",
    version=version,
    author="David Blair",
    author_email="david.blair@ucsf.edu",
    description="A fully generative probability model that maps gene-based knowledge directly to disease symptoms, including their reported frequencies.",
    long_description_content_type="text/markdown",
    url="https://github.com/daverblair/VAE_G2P",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'scipy',
        'torch',
        'pyro-ppl',
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import pandas as pd
import torch
import numpy as np

class TorchDataWrapper(torch.utils.data.Dataset):

    def _collateFunction(self,x):
        return x[0]

    def _indexSplit(self,dataSetSize,totalNumBatches):
        if totalNumBatches>0:
            nEachBatch, extras = divmod(dataSetSize, totalNumBatches)
            section_sizes = ([0] +extras * [nEachBatch+1] +(totalNumBatches-extras) * [nEachBatch])
            return np.array(section_sizes, dtype=np.int32).cumsum()
        else:
            return np.array([0]+[dataSetSize], dtype=np.int32)

    def __init__(self,primary_dataset,batch_size,index='Training'):
        """

        Wrapper for DatasetTrainingState to allow for rapid subset sampling using PyTorch DataLoader, which allows for multi-threaded loading/queueing of data.

        Parameters
        ----------
        training_state : DatasetTrainingState
            DatasetTrainingState to be wrapped
        batch_size : int
            Batch size for the sampler.

        Returns
        -------
        None

        """


        self.primary_dataset = primary_dataset
        assert index in ['Training','Validation','Test'],"Only Training, Validation or Test dataset indices allowed."
        if index=='Training':
            self.sampling_index=self.primary_dataset.training_index
        elif index=='Validation':
            self.sampling_index=self.primary_dataset.validation_index
        else:
            self.sampling_index=self.primary_dataset.testing_index

        np.random.shuffle(self.sampling_index)
        self.totalNumBatches,leftover = divmod(len(self.sampling_index),batch_size)
        self.splits = self._indexSplit(self.sampling_index.shape[0],self.totalNumBatches)
        if self.totalNumBatches == 0:
            self.totalNumBatches+=1

    def __len__(self):
        return self.totalNumBatches

    def shuffle_index(self):
        np.random.shuffle(self.sampling_index)

    def __getitem__(self,index):
        batchIndex=self.sampling_index[self.splits[index]:self.splits[index+1]]
        return self.primary_dataset.ReturnDataArrays(batchIndex)




import collections
import torch
from torch import nn
import math

class FixedWeights_Linear(nn.Module):

    __constants__ = ['bias', 'in_features', 'out_features']

    def __init__(self, in_features, out_features,weight_matrix, bias=True):
        """
        Modifies the Linear neural network class such that the weights are fixed and not updated during inference.
        
        """
        super(FixedWeights_Linear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(weight_matrix.clone().detach().type(torch.float32),requires_grad=False)
        if bias:
            self.bias = nn.Parameter(torch.zeros(self.out_features,dtype=torch.float32), requires_grad=True)
        else:
            self.register_parameter('bias', None)

    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            torch.nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input: torch.Tensor):
        return torch.nn.functional.linear(input, self.weight, self.bias)

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )

class FCLayers(nn.Module):
    """
    A helper class to build fully-connected layers for a neural network. Adapted from scVI (https://github.com/scverse/scvi-tools)
    Parameters
    ----------
    n_in
        The dimensionality of the input
    n_out
        The dimensionality of the output
    n_layers
        The number of fully-connected hidden layers
    n_hidden
        The number of nodes per hidden layer
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    use_batch_norm
        Whether to have `BatchNorm` layers or not
    use_activation
        Whether to have layer activation or not
    activation_fn
        Which activation function to use
    """

    def __init__(
        self,
        n_in: int,
        n_out: int,
        n_layers: int = 1,
        n_hidden: int = 128,
        dropout_rate: float = 0.1,
        use_batch_norm: bool = True,
        use_activation: bool = True,
        activation_fn: nn.Module = nn.ReLU,
    ):
        super().__init__()
        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]

        self.fc_layers = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        f"Layer {i}",
                        nn.Sequential(
                            nn.Linear(
                                n_in ,
                                n_out,
                                bias=True,
                            ),
                            # non-default params come from defaults in original Tensorflow implementation
                            nn.BatchNorm1d(n_out, momentum=0.01, eps=0.001)
                            if use_batch_norm
                            else None,
                            activation_fn() if use_activation else None,
                            nn.Dropout(p=dropout_rate) if dropout_rate > 0 else None,
                        ),
                    )
                    for i, (n_in, n_out) in enumerate(
                        zip(layers_dim[:-1], layers_dim[1:])
                    )
                ]
            )
        )

    def forward(self, x: torch.Tensor):
        for i, layers in enumerate(self.fc_layers):
            for layer in layers:
                if layer is not None:
                        x = layer(x)
        return x


class FC_MeanScaleEncoder(nn.Module):


    def __init__(self, n_input: int, n_output: int, n_layers: int = 2, n_hidden: int = 128, dropout_rate: float = 0.1, use_batch_norm:bool=True,use_activation: bool = True,activation_fn: nn.Module = nn.ReLU):
        """Fully connected neural newtwork used to generate probability distribution over latent space used to encode observed data. 
        
        Parameters
        ----------
        n_input : int
            Dimension of the diagnostic space
        n_output : int
            Dimension of the latent space
        n_layers : int, optional
            Number of hidden layers in network. Default is 2.
        n_hidden : int, optional
            Number of hidden nodes. Default is 128.
        dropout_rate : float, optional
            Dropout rate for regularization during training. Default rate is 0.1. Set to 0.0 to remove dropout from training. 
        use_batch_norm : bool, optional
            Indicates whether to use batch normalization during traning. 
        use_activation : bool, optional
            Indicates whether to use activation function to transform final output of network. Default is True (improves expressivity via non-linearity)
        activation_fn : nn.Module, optional
            Activation function to use. Default is ReLU
        """
        super(FC_MeanScaleEncoder,self).__init__()

        self.encoder = FCLayers(n_in=n_input, n_out=n_hidden, n_layers=n_layers,
                                n_hidden=n_hidden, dropout_rate=dropout_rate,use_batch_norm=use_batch_norm,use_activation=use_activation,activation_fn=activation_fn)
        self.mean_encoder = nn.Linear(n_hidden, n_output,bias=True)
        self.var_encoder = nn.Linear(n_hidden, n_output,bias=True)

    def forward(self, x):
        r"""The forward computation for a single sample.
         #. Encodes the data into latent space using the encoder network
        :param x: tensor with shape (n_input,)
        :param cat_list: list of category membership(s) for this sample
        """

        # Parameters for latent distribution
        q = self.encoder(x)
        q_m = self.mean_encoder(q)
        log_q_v = self.var_encoder(q)+1e-6 #added to prevent underflow errors
        
        return q_m,torch.exp(0.5*log_q_v)




class FC_Decoder(nn.Module):

    def __init__(self, n_input: int, n_output: int, n_layers: int = 2, n_hidden: int = 128, dropout_rate: float = 0.1, use_batch_norm:bool=True,use_activation: bool = True,activation_fn: nn.Module = nn.ReLU):
        """Fully connected neural network used to decode latent phenotype space into logit-probabilities for a set of diagnoses.
        
        Parameters
        ----------
        n_input : int
            Dimension of the latent space
        n_output : int
            Dimension of the diagnostic space
        n_layers : int, optional
            Number of hidden layers in network. Default is 2.
        n_hidden : int, optional
            Number of hidden nodes. Default is 128.
        dropout_rate : float, optional
            Dropout rate for regularization during training. Default rate is 0.1. Set to 0.0 to remove dropout from training. 
        use_batch_norm : bool, optional
            Indicates whether to use batch normalization during traning. 
        use_activation : bool, optional
            Indicates whether to use activation function to transform final output of network. Default is True (improves expressivity via non-linearity)
        activation_fn : nn.Module, optional
            Activation function to use. Default is ReLU
        """
        super(FC_Decoder,self).__init__()
        self.nonlinear_latent = FCLayers(n_in=n_input, n_out=n_hidden, n_layers=n_layers,n_hidden=n_hidden,use_batch_norm=use_batch_norm,use_activation=use_activation,activation_fn=activation_fn)
        self.output_layer =  nn.Linear(n_hidden, n_output,bias=True)


    def forward(self,latent_var):
        return self.output_layer(self.nonlinear_latent(latent_var))


    def _set_bias(self,bias_vec):
        with torch.no_grad():
            self.output_layer.bias[:]=bias_vec[:]

class Linear_Decoder(nn.Module):


    
    def __init__(self,n_input: int, n_output: int):
        """Decodes a latent space into the paramters defining the logit-probabilities for a set of diagnoses.
        
        Parameters
        ----------
        n_input : int
            Dimension of the latent space
        n_output : int
            Dimension of the symptom space
        """
        super(Linear_Decoder,self).__init__()
        self.linear_latent = nn.Linear(n_input,n_output,bias=True)

    def forward(self,latent_var):
        return self.linear_latent(latent_var)

    def _set_bias(self,bias_vec):
        with torch.no_grad():
            self.linear_latent.bias[:]=bias_vec[:]



class Fixed_Linear_Decoder(nn.Module):
    def __init__(self, decoding_matrix):
        """Decodes a latent space into the paramters defining the logit-probabilities for a set of diagnoses. The dimension of the diagnosis space is determined from the decoding_matrix provided by the user (dim=1). Note, the decoding matrix is NOT updated during inference. 
        
        Parameters
        ----------
        decoding_matrix : np.array or torch.tensor

        Matrix used to decode a latent variable space (dim=0) into a set of logit parameters defining diagnostic probabilities. 
        
        """
        super(Fixed_Linear_Decoder,self).__init__()
        self.linear_latent=FixedWeights_Linear(decoding_matrix.shape[1],decoding_matrix.shape[0],decoding_matrix,bias=True)

    def _set_bias(self,bias_vec):
        with torch.no_grad():
            self.linear_latent.bias[:]=bias_vec[:]

    def forward(self,latent_var):
        return self.linear_latent(latent_var)











import pandas as pd
import numpy as np
import torch
import pickle
import pyro
import copy

from .cvae_model import DiseaseCondVAE
from .basic_g2p import BasicG2PModel
from .data import GeneToPhenotypeDataset
from .optim import AutoEncoderOptimizer

__version__ = "0.0.5"


class VAE_G2P:

	def _readModelFromFile(self,fName):
		with open(fName,'rb') as f:
			model_dict = torch.load(f,map_location='cpu')
		return model_dict


	def __init__(self,diseaseGeneDataset,nLatentDim,isLinear=True,**kwargs):
		self.diseaseGeneDataset = diseaseGeneDataset
		self.nLatentDim=nLatentDim
		self.isLinear=isLinear
		self.numSymptoms=self.diseaseGeneDataset.num_symptoms
		self.numFreqCats=self.diseaseGeneDataset.num_ordinal_freqs

		self.all_model_kwargs = kwargs

		if 'encoder_hyperparameters' not in self.all_model_kwargs.keys():
			self.all_model_kwargs['encoder_hyperparameters']={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.1, 'use_batch_norm':True}
		if 'decoder_hyperparameters' not in self.all_model_kwargs.keys():
			self.all_model_kwargs['decoder_hyperparameters']={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.1, 'use_batch_norm':True}

		if 'missing_freq_priors' not in self.all_model_kwargs.keys():
			self.all_model_kwargs['missing_freq_priors']=[0.0,3.0]

		if 'frequency_cut_points' not in self.all_model_kwargs.keys():
			self.all_model_kwargs['frequency_cut_points']=[0.04,0.3,0.8,0.99]

		assert (len(self.all_model_kwargs['frequency_cut_points'])+1)==self.diseaseGeneDataset.num_ordinal_freqs, "Number of model cutpoints must equal number frequency classes in dataset, excluding the Unknown class"

	
		self.basic_g2p=BasicG2PModel(self.diseaseGeneDataset,network_hyperparameters=self.all_model_kwargs['decoder_hyperparameters'],cut_points=self.all_model_kwargs['frequency_cut_points'])

		self.vae_g2p_model=DiseaseCondVAE(self.numSymptoms,self.numFreqCats,self.nLatentDim,self.basic_g2p,isLinear=self.isLinear,encoder_hyperparameters=self.all_model_kwargs['encoder_hyperparameters'],decoder_hyperparameters=self.all_model_kwargs['decoder_hyperparameters'],missing_freq_prior_mean=self.all_model_kwargs['missing_freq_priors'][0],missing_freq_prior_scale=self.all_model_kwargs['missing_freq_priors'][1],cut_points=self.all_model_kwargs['frequency_cut_points'])


	def Fit(self,batch_size,logFile=None,verbose=True,monitor_validation=True,prior_model_state=None,**kwargs):
		"""


		Parameters
		----------


		batch_size : int,
		    Size of dataset batches for inference. 

		verbose : bool, optional
		    Indicates whether or not to print (to std out) the loss function values and error after every epoch. The default is True.

		logFile: str, optional

		    File to log model fitting process.

		Keyword Parameters
		----------
		learningRate: float, optional
		    Specifies the maximum learning rate used during inference. Default is 0.05

		errorTol: float, optional
		    Error tolerance in ELBO (computed on held out validation data) to determine convergence. Default is 1e-4.

		numParticles: int, optional
		    Number of particles (ie random samples) used to approximate gradient. Default is 1. Computational cost increases linearly with value.

		maxEpochs: int, optional
		    Maximum number of epochs (passes through training data) for inference. Note, because annealing and learning rate updates depend on maxEpochs, this offers a simple way to adjust the speed at which these values change. Default is 200.

		computeDevice: str or None, optional
		    Specifies compute device for inference. Default is None, which instructs algorithm to use cpu. Two other options are supported: 'cuda' and 'mps'. Note, if device number is not included (ex: 'cuda:0'), then it is automatically assumed to be '0'

		numDataLoaders: int
		    Specifies the number of threads used to process data and prepare for upload into the gpu. Note, due to the speed of gpu, inference can become limited by data transfer speed, hence the use of multiple DataLoaders to improve this bottleneck. Default is 0, meaning just the dedicated cpu performs data transfer.

		KLAnnealingParams: dict with keys 'initialTemp','maxTemp','fractionalDuration','schedule'
		    Parameters that define KL-Annealing strategy used during inference, important for avoiding local optima. Note, annealing is only used for computation of ELBO and gradients on training data. Validation data ELBO evaluation, used to monitor convergence, is performed at the maximum desired temperature (typically 1.0, equivalent to standard variational inference). Therefore, it is possible for the model to converge even when the temperature hasn't reached it's final value. It's also possible that further cooling would find a better optimum, but this is highly unlikely in practice.
		    initialTemp--initial temperature during inference. Default: 1.0 (no annealing)
		    maxTemp--final temperature obtained during inference. Default: 1.0 (standard variational inference)
		    fractionalDuration--fraction of inference epochs used for annealing. Default is 0.25
		    schedule--function used to change temperature during inference. Defualt is 'cosine'. Options: 'cosine','linear'



		Returns
		-------
		output : list
		    List containing the following information: [loss function value of best model (computed on validation data),sequence of training loss values, sequence of validation loss values, error estimates across iterations (computed on validation data)].



		"""


		######### Parse Keyword Arguments #########
		allKeywordArgs = list(kwargs.keys())


		if 'learningRate' in allKeywordArgs:
		    learningRate=kwargs['learningRate']
		else:
		    learningRate=0.01


		if 'errorTol' in allKeywordArgs:
		    errorTol=kwargs['errorTol']
		else:
		    errorTol=1e-4

		if 'numParticles' in allKeywordArgs:
		    numParticles=kwargs['numParticles']
		else:
		    numParticles=1


		if 'maxEpochs' in allKeywordArgs:
		    maxEpochs=kwargs['maxEpochs']
		else:
		    maxEpochs=500


		if 'computeDevice' in allKeywordArgs:
		    computeDevice=kwargs['computeDevice']
		else:
		    computeDevice=None

		if 'numDataLoaders' in allKeywordArgs:
		    numDataLoaders=kwargs['numDataLoaders']
		    if computeDevice in [None,'cpu']:
		        assert numDataLoaders==0,"Specifying number of dataloaders other than 0 only relevant when using GPU computing"
		else:
		    numDataLoaders=0

		if 'KLAnnealingParams' in allKeywordArgs:
		    KLAnnealingParams=kwargs['KLAnnealingParams']
		    assert set(KLAnnealingParams.keys())==set(['initialTemp','maxTemp','fractionalDuration','schedule']),"KL Annealing Parameters must be dictionary with the following keys: 'initialTemp','maxTemp','fractionalDuration','schedule'"
		else:
		    KLAnnealingParams={'initialTemp':1.0,'maxTemp':1.0,'fractionalDuration':1.0,'schedule': 'cosine'}


		if 'EarlyStopPatience' in allKeywordArgs:
			EarlyStopPatience=kwargs['EarlyStopPatience']
		else:
			EarlyStopPatience=10


		pyro.clear_param_store()
		if prior_model_state is None:
			self.basic_g2p.Fit(batch_size,learningRate,maxEpochs,errorTol,compute_device=computeDevice,numDataLoaders=numDataLoaders,early_stop_patience=EarlyStopPatience,monitor_validation=False)
		else:
			self.basic_g2p.LoadModel(prior_model_state)


		optimizer=AutoEncoderOptimizer(self.vae_g2p_model,self.diseaseGeneDataset,optimizationParameters={'learningRate': learningRate,'maxEpochs': maxEpochs,'numParticles':numParticles},computeConfiguration={'device':computeDevice,'numDataLoaders':numDataLoaders},KLAnnealingParams=KLAnnealingParams)
		output=optimizer.BatchTrain(batch_size,errorTol=errorTol,verbose=verbose,logFile=logFile,monitor_validation=monitor_validation,early_stop_patience=EarlyStopPatience)
		return output

	def PredictEmbedFromGeneOnly(self,gene_list,returnStdErrors=False):
		assert len(set(gene_list).difference(self.diseaseGeneDataset.gene_map.keys()))==0,"The following genes are not in the embedding table: {0:s}".format(','.join(list(set(gene_vec).difference(self.diseaseGeneDataset.gene_map.keys()))))

		gene_data=self.diseaseGeneDataset.ReturnGeneDataArrays(gene_list)
		modes_for_prior = torch.cat(self.vae_g2p_model.baseline_g2p_model.basic_net(gene_data),axis=1)
		p_m,p_std = self.vae_g2p_model.prior_model(torch.cat((modes_for_prior,gene_data),axis=1))

		output_table={'Genes':gene_list,'Embeddings':[x for x in p_m.detach().numpy()]}

		if returnStdErrors:
			output_table['Std Errors']=[x for x in p_std.detach().numpy()]
		output_table=pd.DataFrame(output_table)
		output_table.set_index('Genes',inplace=True)
		return output_table


	def PredictSymptomsFromGeneOnly(self,gene_list,numSamples=0):
		embeds=self.PredictEmbedFromGeneOnly(gene_list,returnStdErrors=True)
		z_loc=np.vstack(embeds.loc[gene_list]['Embeddings'].values)
		z_scale=np.vstack(embeds.loc[gene_list]['Std Errors'].values)
		if numSamples==0:
			symptom_probs=torch.sigmoid(self.vae_g2p_model.symptom_annotation_decoder.forward(torch.tensor(z_loc))).detach().numpy()

			symptom_freqs=torch.exp(pyro.distributions.OrderedLogistic(self.vae_g2p_model.symptom_frequency_decoder(torch.tensor(z_loc)),self.vae_g2p_model.cut_points).logits).detach().numpy()
		else:
			samples_from_priors = torch.normal(torch.tensor(z_loc)*torch.ones((numSamples,z_loc.shape[0],z_loc.shape[1])),torch.tensor(z_scale)*torch.ones((numSamples,z_loc.shape[0],z_loc.shape[1])))

			symptom_probs=np.zeros((len(gene_list),self.numSymptoms))
			symptom_freqs=np.zeros((len(gene_list),self.numSymptoms,self.numFreqCats))

			for i in range(numSamples):
				symptom_probs+=torch.sigmoid(self.vae_g2p_model.symptom_annotation_decoder.forward(samples_from_priors[i])).detach().numpy()/numSamples
				symptom_freqs+=torch.exp(pyro.distributions.OrderedLogistic(self.vae_g2p_model.symptom_frequency_decoder(samples_from_priors[i]),self.vae_g2p_model.cut_points).logits).detach().numpy()/numSamples

		output_table=pd.DataFrame([],index=self.diseaseGeneDataset.symptom_map.keys(),columns=gene_list)
		for i,gene in enumerate(gene_list):
			output_table[gene]=list(zip(symptom_probs[i],symptom_freqs[i]))
		return output_table

	

	def _per_datam_elbo(self,hpo_freq_pairs,gene_list,num_particles=10):
		assert isinstance(hpo_freq_pairs,list),"The provided HPO-frequncy pairs must be an interable of ('HPO Symptom','Frequency') nested within a list."

		assert isinstance(gene_list,list),"Conditional VAE model expects a list of genes in order to compute ELBO"
		assert len(hpo_freq_pairs)==len(gene_list) or (len(hpo_freq_pairs)==1 or len(gene_list)==1),"Number of provided symptom-frequency pair datasets must equal number of genes, unless number of genes or the number of symptom-frequency pair datasets is equal to 1, at which point the data type with length 1 is broadcasted to the same size as the other dataset."

		#generate data tensors, allows broadcasting 
		annot_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.numSymptoms),dtype=torch.float32)
		freq_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.numSymptoms),dtype=torch.float32)

		if len(gene_list)<len(hpo_freq_pairs):
			gene_list=gene_list*len(hpo_freq_pairs)
		elif len(gene_list)>len(hpo_freq_pairs):
			hpo_freq_pairs=hpo_freq_pairs*len(gene_list)

		for t,hpo_freq_vec in enumerate(hpo_freq_pairs):
			idx_vec=[self.diseaseGeneDataset.symptom_map[x[0]] for x in hpo_freq_vec]
			freq_vec=[self.diseaseGeneDataset.ordinal_freq_map[x[1]] for x in hpo_freq_vec]
			annot_array[t,idx_vec]=1.0
			freq_array[t,idx_vec]=torch.tensor(freq_vec,dtype=torch.float32)

		data=(annot_array,freq_array,self.diseaseGeneDataset.ReturnGeneDataArrays(gene_list))

		elbo=self.vae_g2p_model.per_datum_ELBO(*data,num_particles=num_particles)
		return elbo.detach().numpy().ravel()

	def PerplexCompare(self,hpo_freq_pairs,gene_list,index=None,num_particles=10):
		perplex_vec = -1.0*self._per_datam_elbo(hpo_freq_pairs,gene_list,num_particles=num_particles)
		model_compare_results=self.basic_g2p.PerplexCompareNull(hpo_freq_pairs,gene_list)
		model_compare_results['VAE_G2P Perplex.']=perplex_vec
		model_compare_results['VAE_G2P-BasicG2PModel']=model_compare_results['VAE_G2P Perplex.']-model_compare_results['BasicG2PModel Perplex.']
		model_compare_results['VAE_G2P-Null']=model_compare_results['VAE_G2P Perplex.']-model_compare_results['Null Perplex.']
		model_compare_results['Genes']=gene_list
		model_compare_results=pd.DataFrame(model_compare_results)
		if index is not None:
			model_compare_results['Index']=index
		else:
			model_compare_results['Index']=['Dis_{0:d}'.format(x) for x in range(len(gene_list))]
		model_compare_results.set_index('Index',inplace=True)
		return model_compare_results


	def EmbedDisease(self,hpo_freq_pairs,gene_list,index=None,returnStdErrors=False):
		assert isinstance(hpo_freq_pairs,list),"The provided HPO-frequncy pairs must be an interable of ('HPO Symptom','Frequenc') nested within a list."
		assert isinstance(gene_list,list),"Conditional VAE model expects a list of genes in order to compute ELBO"
		assert len(hpo_freq_pairs)==len(gene_list) or (len(hpo_freq_pairs)==1 or len(gene_list)==1),"Number of provided symptom-frequency pair datasets must equal number of genes, unless number of genes or the number of symptom-frequency pair datasets is equal to 1, at which point the data type with length 1 is broadcasted to the same size as the other dataset."

		#generate data tensors, allows broadcasting 
		annot_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.numSymptoms),dtype=torch.float32)
		freq_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.numSymptoms),dtype=torch.float32)

		if len(gene_list)<len(hpo_freq_pairs):
			gene_list=gene_list*len(hpo_freq_pairs)
		elif len(gene_list)>len(hpo_freq_pairs):
			hpo_freq_pairs=hpo_freq_pairs*len(gene_list)

		for t,hpo_freq_vec in enumerate(hpo_freq_pairs):
			idx_vec=[self.diseaseGeneDataset.symptom_map[x[0]] for x in hpo_freq_vec]
			freq_vec=[self.diseaseGeneDataset.ordinal_freq_map[x[1]] for x in hpo_freq_vec]
			annot_array[t,idx_vec]=1.0
			freq_array[t,idx_vec]=torch.tensor(freq_vec,dtype=torch.float32)

		data=(annot_array,freq_array,self.diseaseGeneDataset.ReturnGeneDataArrays(gene_list))
		p_m,p_std=self.vae_g2p_model.posterior_latent_state(*data)

		if index is None:
			index=['Dis_{0:d}'.format(x) for x in range(len(gene_list))]
		output_table={'Index':index,'Gene':gene_list,'Embeddings':[x for x in p_m.detach().numpy()]}

		if returnStdErrors:
			output_table['Std Errors']=[x for x in p_std.detach().numpy()]
		output_table=pd.DataFrame(output_table)
		output_table.set_index('Index',inplace=True)
		return output_table

	def _embedDisease(self,hpo_freq_pairs,gene_list,returnStdErrors=False):
		assert isinstance(hpo_freq_pairs,list),"The provided HPO-frequncy pairs must be an interable of ('HPO Symptom','Frequenc') nested within a list."
		assert isinstance(gene_list,list),"Conditional VAE model expects a list of genes in order to compute ELBO"
		assert len(hpo_freq_pairs)==len(gene_list) or (len(hpo_freq_pairs)==1 or len(gene_list)==1),"Number of provided symptom-frequency pair datasets must equal number of genes, unless number of genes or the number of symptom-frequency pair datasets is equal to 1, at which point the data type with length 1 is broadcasted to the same size as the other dataset."

		#generate data tensors, allows broadcasting 
		annot_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.numSymptoms),dtype=torch.float32)
		freq_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.numSymptoms),dtype=torch.float32)

		if len(gene_list)<len(hpo_freq_pairs):
			gene_list=gene_list*len(hpo_freq_pairs)
		elif len(gene_list)>len(hpo_freq_pairs):
			hpo_freq_pairs=hpo_freq_pairs*len(gene_list)

		for t,hpo_freq_vec in enumerate(hpo_freq_pairs):
			idx_vec=[self.diseaseGeneDataset.symptom_map[x[0]] for x in hpo_freq_vec]
			freq_vec=[self.diseaseGeneDataset.ordinal_freq_map[x[1]] for x in hpo_freq_vec]
			annot_array[t,idx_vec]=1.0
			freq_array[t,idx_vec]=torch.tensor(freq_vec,dtype=torch.float32)

		data=(annot_array,freq_array,self.diseaseGeneDataset.ReturnGeneDataArrays(gene_list))
		p_m,p_std=self.vae_g2p_model.posterior_latent_state(*data)

		if returnStdErrors:
			return p_m.detach().numpy(),p_std.detach().numpy()
		else:
			return p_m.detach().numpy()


	def EstimateMissingAnnotationRates(self,hpo_freq_pairs,gene_list,index=None):
		assert len(hpo_freq_pairs)==len(gene_list),"Expect length of HPO-Freqency Pair sets and gene lists to be the same."
		mapping_dict={}
		for i,hpo_freq_vec in enumerate(hpo_freq_pairs):
			mapping_dict[gene_list[i]]=torch.tensor([self.diseaseGeneDataset.symptom_map[x[0]] for x in hpo_freq_vec],dtype=torch.long)

		post_mean=self._embedDisease(hpo_freq_pairs,gene_list=gene_list,returnStdErrors=False)

		if index is None:
			index=['Dis_{0:d}'.format(x) for x in range(len(gene_list))]
		output_table={'Index':index,'Gene':gene_list,'Missing Annot. Rates':[]}
		for i,post_vec in enumerate(post_mean):
			embed_tensor=torch.tensor(post_vec,dtype=torch.float32)
			pred_missing_full=self.vae_g2p_model.missing_freq_disease_decoder(embed_tensor)+self.vae_g2p_model.missing_freq_intercepts_post_mean
			pred_missing=torch.sigmoid(pred_missing_full[mapping_dict[gene_list[i]]]).detach().numpy()
			output_table['Missing Annot. Rates']+=[list(zip([self.diseaseGeneDataset.inverse_symptom_map[x] for x in mapping_dict[gene_list[i]].detach().numpy()],pred_missing))]
		output_table=pd.DataFrame(output_table)
		output_table.set_index('Index',inplace=True)
		return output_table


	def ImputeMissingSymptomFrequencies(self,hpo_freq_pairs,gene_list,index=None):
		assert len(hpo_freq_pairs)==len(gene_list),"Expect length of HPO-Freqency Pair sets and gene lists to be the same."
		mapping_dict={}
		for i,hpo_freq_vec in enumerate(hpo_freq_pairs):
			mapping_dict[gene_list[i]]=torch.tensor([self.diseaseGeneDataset.symptom_map[x[0]] for x in hpo_freq_vec if x[1]==self.diseaseGeneDataset.missing_label],dtype=torch.long)

		post_mean=self._embedDisease(hpo_freq_pairs,gene_list,returnStdErrors=False)


		if index is None:
			index=['Dis_{0:d}'.format(x) for x in range(len(gene_list))]
		output_table={'Index':index,'Gene':gene_list,'Missing Freqencies':[]}
		for i,post_vec in enumerate(post_mean):
			embed_tensor=torch.tensor(post_vec,dtype=torch.float32)
			pred_class_full=pyro.distributions.OrderedLogistic(self.vae_g2p_model.symptom_frequency_decoder(embed_tensor.unsqueeze(dim=0)),self.vae_g2p_model.cut_points).logits.squeeze(dim=0)
			pred_annots=torch.exp(pred_class_full[mapping_dict[gene_list[i]],:]).detach().numpy()
			output_table['Missing Freqencies']+=[list(zip([self.diseaseGeneDataset.inverse_symptom_map[x] for x in mapping_dict[gene_list[i]].detach().numpy()],pred_annots))]
		output_table=pd.DataFrame(output_table)
		output_table.set_index('Index',inplace=True)
		return output_table

	def SuggestNewSymptoms(self,hpo_freq_pairs,gene_list,annot_rate_threshold=0.0,index=None):
		assert len(hpo_freq_pairs)==len(gene_list),"Expect length of HPO-Freqency Pair sets and gene lists to be the same."

		mapping_dict={}
		for i,hpo_freq_vec in enumerate(hpo_freq_pairs):
			mapping_dict[gene_list[i]]=set([x[0] for x in hpo_freq_vec])

		post_mean=self._embedDisease(hpo_freq_pairs,gene_list,returnStdErrors=False)

		if index is None:
			index=['Dis_{0:d}'.format(x) for x in range(len(gene_list))]
		output_table={'Index':index,'Gene':gene_list,'New Symptoms':[]}
		for i,post_vec in enumerate(post_mean):
			embed_tensor=torch.tensor(post_vec,dtype=torch.float32)
			pred_symptoms=torch.sigmoid(self.vae_g2p_model.symptom_annotation_decoder(embed_tensor.unsqueeze(dim=0)).squeeze(dim=0)).detach().numpy()
			pred_class_full=pyro.distributions.OrderedLogistic(self.vae_g2p_model.symptom_frequency_decoder(embed_tensor.unsqueeze(dim=0)),self.vae_g2p_model.cut_points).logits.squeeze(dim=0)
			pred_class_full=torch.exp(pred_class_full).detach().numpy()

			allowed_symptoms=np.arange(self.numSymptoms)
			allowed_symptoms=allowed_symptoms[pred_symptoms>=annot_rate_threshold]
			pred_class_full=pred_class_full[pred_symptoms>=annot_rate_threshold]
			pred_symptoms=pred_symptoms[pred_symptoms>=annot_rate_threshold]

			
			allowed_symptoms=allowed_symptoms[np.argsort(pred_symptoms)[::-1]]
			pred_class_full=pred_class_full[np.argsort(pred_symptoms)[::-1]]
			pred_symptoms=np.sort(pred_symptoms)[::-1]
			output_vec=[]

			for j in range(allowed_symptoms.shape[0]):
				symptom=self.diseaseGeneDataset.inverse_symptom_map[allowed_symptoms[j]]
				if symptom not in mapping_dict[gene_list[i]]:
					output_vec+=[(symptom,pred_symptoms[j],pred_class_full[j])]
			output_table['New Symptoms']+=[output_vec]

		output_table=pd.DataFrame(output_table)
		output_table.set_index('Index',inplace=True)
		return output_table




	def LoadModel(self,stored_model):
		"""
		Loads previously fit model either from a dictionary (generated using PackageModel) or from a file path (with file constructed using PackageModel)

		Parameters
		----------
		stored_model : either dict or str (file path)

		Returns
		-------
		None.

		"""
		if not isinstance(stored_model,dict):
			assert isinstance(stored_model,str),"Expects file name if not provided with dictionary."
			stored_model = self._readModelFromFile(stored_model)

		assert set(stored_model.keys())==set(['model_state','meta_data','variational_post_params','baseline_g2p_model']),"Model dictionary must contain the following elements: 'model_state','meta_data','baseline_g2p_model'"
		self.basic_g2p=BasicG2PModel(self.diseaseGeneDataset,network_hyperparameters=stored_model['meta_data']['all_model_kwargs']['decoder_hyperparameters'],cut_points=stored_model['meta_data']['all_model_kwargs']['frequency_cut_points'])
		self.vae_g2p_model=DiseaseCondVAE(
			stored_model['meta_data']['numSymptoms'],
			stored_model['meta_data']['numFreqCats'],
			stored_model['meta_data']['nLatentDim'],
			self.basic_g2p,
			isLinear=stored_model['meta_data']['isLinear'],
			encoder_hyperparameters=stored_model['meta_data']['all_model_kwargs']['encoder_hyperparameters'],
			decoder_hyperparameters=stored_model['meta_data']['all_model_kwargs']['decoder_hyperparameters'],
			missing_freq_prior_mean=stored_model['meta_data']['all_model_kwargs']['missing_freq_priors'][0],
			missing_freq_prior_scale=stored_model['meta_data']['all_model_kwargs']['missing_freq_priors'][1],
			cut_points=stored_model['meta_data']['all_model_kwargs']['frequency_cut_points']
			)
		self.vae_g2p_model.load_state(stored_model)
	


	def SaveModel(self,fName=None):
		"""
		Packages the current model and returns it as a python dictionary. Will optionally write this dictionary to disk using PyTorch.

		Parameters
		----------
		fName : str, default None
		    File path to save model to disk. The default is None, which means that only a model dictionary will be returned.

		Returns
		-------
		model_dict : dict
		    Dictionary containing fitted model parameters in addition to general meta data.

		"""
		model_dict = self.vae_g2p_model.package_state()
		model_dict['meta_data']={}
		model_dict['meta_data']['numSymptoms']=self.numSymptoms
		model_dict['meta_data']['numFreqCats']=self.numFreqCats
		model_dict['meta_data']['nLatentDim']=self.nLatentDim
		model_dict['meta_data']['isLinear']=self.isLinear
		model_dict['meta_data']['all_model_kwargs']=self.all_model_kwargs
		if fName is not None:
		    with open(fName,'wb') as f:
		        torch.save(model_dict,f)
		return model_dict




import numpy as np
import torch
from torch import nn
import pyro
import pyro.distributions as dist
from torch.utils import data
from collections import deque
from scipy.stats import linregress
import copy

from .networks import FCLayers
from .data import GeneToPhenotypeDataset
from .data_wrapper import TorchDataWrapper




def _rel_diff(curr,prev):
    return abs((curr - prev) / prev)

def _collate_it(x):
    return x[0]


def _returnELBOSlopeError(elbo_deque):
    """ Takes that average of the deque containing ELBO values. Helps determine convergence in the face of noisy SVI by taking the slope of the ELBO values as a function of iteration number. Error is computed as abs(slope)/mean(ELBO)
    
    Parameters
    ----------
    elbo_deque : collections.deque
        Deque containing previous N ELBO values (depends on deque size).
    
    Returns
    -------
    float
        ELBO error 
    """
    return abs(linregress(range(len(elbo_deque)),elbo_deque)[0]/np.mean(elbo_deque))



class BasicG2PMapLoss(nn.Module):
    def __init__(self):
        super(BasicG2PMapLoss,self).__init__()

    def forward(self,symptoms_logits,missing_logits, freq_logits, obs_symptoms, obs_freq_classes,cut_points):
        symptom_annot_loss=dist.Bernoulli(logits=symptoms_logits).log_prob(obs_symptoms).sum()

        freq_missing=torch.zeros(obs_freq_classes.shape,dtype=torch.float32,device=freq_logits.device)
        freq_missing[(obs_freq_classes==-1.0)]=1.0
        missing_loss=dist.Bernoulli(logits=missing_logits[obs_symptoms==1.0]).log_prob(freq_missing[obs_symptoms==1.0]).sum()

        
        freq_missing_mask = torch.zeros(obs_freq_classes.shape,dtype=torch.bool,device=freq_logits.device)
        freq_obs_mask = torch.zeros(obs_freq_classes.shape,dtype=torch.bool,device=freq_logits.device)
        freq_missing_mask[obs_symptoms==1.0]=1
        freq_obs_mask[(obs_symptoms==1.0)&(obs_freq_classes!=-1.0)]=1
        if freq_obs_mask.sum()>0:
            freq_loss=dist.OrderedLogistic(freq_logits[freq_obs_mask],cut_points).log_prob(obs_freq_classes[freq_obs_mask]).sum()
        else:
            freq_loss=0.0

        return -1.0*(symptom_annot_loss+missing_loss+freq_loss)


class BasicG2PMapLoss_Categortical(nn.Module):
    def __init__(self):
        super(BasicG2PMapLoss_Categortical,self).__init__()

    def forward(self,symptoms_logits,missing_logits, freq_logits, obs_symptoms, obs_freq_classes):
        symptom_annot_loss=dist.Bernoulli(logits=symptoms_logits).log_prob(obs_symptoms).sum()

        freq_missing=torch.zeros(obs_freq_classes.shape,dtype=torch.float32,device=freq_logits.device)
        freq_missing[(obs_freq_classes==-1.0)]=1.0
        missing_loss=dist.Bernoulli(logits=missing_logits[obs_symptoms==1.0]).log_prob(freq_missing[obs_symptoms==1.0]).sum()

        
        freq_missing_mask = torch.zeros(obs_freq_classes.shape,dtype=torch.bool,device=freq_logits.device)
        freq_obs_mask = torch.zeros(obs_freq_classes.shape,dtype=torch.bool,device=freq_logits.device)
        freq_missing_mask[obs_symptoms==1.0]=1
        freq_obs_mask[(obs_symptoms==1.0)&(obs_freq_classes!=-1.0)]=1
        if freq_obs_mask.sum()>0:
            freq_loss=dist.Categorical(logits=freq_logits[freq_obs_mask,:]).log_prob(obs_freq_classes[freq_obs_mask]).sum()
        else:
            freq_loss=0.0

        return -1.0*(symptom_annot_loss+missing_loss+freq_loss)

class BasicG2PMap(nn.Module):
    def __init__(self,
        n_predictors,
        n_symptoms,
        n_freq_classes,
        network_hyperparameters={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.1, 'use_batch_norm':True},
        update_embedding=False,
        cut_points = [0.04,0.3,0.8,0.99]
        ):
        super(BasicG2PMap,self).__init__()
        self.n_predictors=n_predictors
        self.n_outcomes=n_symptoms
        self.n_freq_classes=n_freq_classes
        self.network_hyperparameters=network_hyperparameters
        self.cut_points=torch.logit(torch.tensor(cut_points,dtype=torch.float32))

        self.mlp=FCLayers(n_in=self.n_predictors, n_out=network_hyperparameters['n_hidden'], n_layers=network_hyperparameters['n_layers'],
                                n_hidden=network_hyperparameters['n_hidden'], dropout_rate=network_hyperparameters['dropout_rate'],use_batch_norm=network_hyperparameters['use_batch_norm'],use_activation=True,activation_fn=nn.ReLU)
        self.symptom_layer=nn.Linear(network_hyperparameters['n_hidden'],self.n_outcomes,bias=True)
        self.missing_layer=nn.Linear(network_hyperparameters['n_hidden'],self.n_outcomes,bias=True)
        self.freq_layer=nn.Linear(network_hyperparameters['n_hidden'],self.n_outcomes,bias=True)

    def forward(self, data_arrays):
        hidden_state = self.mlp(data_arrays)
        symptom_logits = self.symptom_layer(hidden_state)
        missing_logits = self.missing_layer(hidden_state)
        freq_logits =  self.freq_layer(hidden_state)
        return symptom_logits,missing_logits,freq_logits

    def package_state(self):
        """
        Packages the model state dict as a dictionary.

        Returns
        -------
        dict
            Model state dict.

        """
        packaged_model_state={}
        packaged_model_state['model_state'] = copy.deepcopy(self.state_dict(keep_vars=True))
        return packaged_model_state

    def load_state(self,prior_model_state):
        """
        Loads model state from dictionary

        Parameters
        ----------
        prior_model_state : dict
            Dictionary of model state produced by PackageCurrentState

        Returns
        -------
        None

        """
        self.load_state_dict(prior_model_state['model_state'],strict=True)

class BasicG2PModel:

    def __init__(self,
        g2p_dataset,
        network_hyperparameters={'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.1, 'use_batch_norm':True},
        cut_points=[0.04,0.3,0.8,0.99]
        ):
        self.g2p_dataset=g2p_dataset
        self.cut_points=torch.logit(torch.tensor(cut_points,dtype=torch.float32))

        self.n_pred=len(self.g2p_dataset.symptom_map)
        self.n_freq_class=len(self.g2p_dataset.ordinal_freq_map)-1
        self.basic_net = BasicG2PMap(self.g2p_dataset.aux_gene_info_table['return_dim'].sum(), self.n_pred,self.n_freq_class,network_hyperparameters=network_hyperparameters)
        self.basic_net.eval()

    def Fit(self,batch_size,learning_rate,max_epochs,errorTol,compute_device=None,numDataLoaders=0,verbose=True,logFile=None,early_stop_patience=5,monitor_validation=False,weight_decay=0.0001):

        if compute_device is None:
            compute_device='cpu'
        else:
            try:
                device_type,device_num=compute_device.split(':')
            except ValueError:
                device_type=compute_device
                device_num='0'
                compute_device=device_type+':'+device_num
            assert device_type in ['cuda','mps','cpu']," Only 'cuda','mps', and 'cpu' supported"
            if device_type=='mps':
                assert torch.backends.mps.is_available()==True,"mps device not available"
            if device_type=='cuda':
                assert torch.cuda.is_available()==True,"cuda device is not available"


        errorComputationWindow = max(int(0.01*max_epochs),2)

        error_window = deque(maxlen=errorComputationWindow)
        elbo_window = deque(maxlen=errorComputationWindow)
        errorVec = []
        trainLoss = []
        validationLoss = []


        if self.g2p_dataset.training_index is not None:
            numTotalTrainingSamples=self.g2p_dataset.training_index.shape[0]
            numTotalValidationSamples=self.g2p_dataset.validation_index.shape[0]
        else:
            raise ValueError("Training state for OMIMDataset.DatasetTrainingState not set prior to calling optimizer. Please set training set using data.DatasetTrainingState.SetNewTrainingState") 

        assert batch_size < numTotalTrainingSamples, "Batch size is greater than or equal to training data size."

        self.basic_net.to(compute_device)
        self.cut_points=self.cut_points.detach().to(compute_device)

        torchTrainingData=TorchDataWrapper(self.g2p_dataset,batch_size,index='Training')
        torchValidationData=TorchDataWrapper(self.g2p_dataset,batch_size,index='Validation')


        optimizer = torch.optim.AdamW(self.basic_net.parameters(), lr=learning_rate,weight_decay=weight_decay)
        criterion = BasicG2PMapLoss()
        criterion.to(compute_device)
        

        bestModelState=self.basic_net.package_state()

        trainingDataLoader = data.DataLoader(torchTrainingData,num_workers=numDataLoaders,collate_fn=_collate_it)
        validationDataLoader = data.DataLoader(torchValidationData,num_workers=numDataLoaders,collate_fn=_collate_it)

        avg_epoch_train_loss = 0.0
        self.basic_net.eval()
        for i,data_batch in enumerate(trainingDataLoader):
            if compute_device.split(':')[0]!='cpu':
                data_batch=tuple([x.to(compute_device) for x in data_batch])
            preds=self.basic_net(data_batch[2])
            loss = criterion(*preds, data_batch[0],data_batch[1],self.cut_points) / (data_batch[0].shape[0]/numTotalTrainingSamples)
            avg_epoch_train_loss+=loss.item()
        avg_epoch_train_loss=avg_epoch_train_loss/(i+1)
        prev_train_loss=avg_epoch_train_loss
        elbo_window.append(prev_train_loss)


        

        avg_epoch_val_loss = 0.0
        for i,data_batch in enumerate(validationDataLoader):
            if compute_device.split(':')[0]!='cpu':
                data_batch=tuple([x.to(compute_device) for x in data_batch])
            preds=self.basic_net(data_batch[2])
            loss = criterion(*preds, data_batch[0],data_batch[1],self.cut_points) / (data_batch[0].shape[0]/numTotalValidationSamples)

            avg_epoch_val_loss+=loss.item()
        avg_epoch_val_loss=avg_epoch_val_loss/(i+1.0)
        bestModelScore=np.inf

        early_stop_count=0

        for epoch in range(max_epochs):
            torchTrainingData.shuffle_index()
            avg_epoch_train_loss = 0.0
            self.basic_net.train()
            for i,data_batch in enumerate(trainingDataLoader):
                if compute_device.split(':')[0]!='cpu':
                    data_batch=tuple([x.to(compute_device) for x in data_batch])
                optimizer.zero_grad()
                with torch.set_grad_enabled(True):
                    preds=self.basic_net(data_batch[2])
                    loss = criterion(*preds, data_batch[0],data_batch[1],self.cut_points) / (data_batch[0].shape[0]/numTotalTrainingSamples)
                    loss.backward()
                    optimizer.step()
                avg_epoch_train_loss+=loss.item()
            avg_epoch_train_loss=avg_epoch_train_loss/(i+1)

            self.basic_net.eval()
            avg_epoch_val_loss = 0.0
            for i,data_batch in enumerate(validationDataLoader):
                if compute_device.split(':')[0]!='cpu':
                    data_batch=tuple([x.to(compute_device) for x in data_batch])
                preds=self.basic_net(data_batch[2])
                loss = criterion(*preds, data_batch[0],data_batch[1],self.cut_points) / (data_batch[0].shape[0]/numTotalValidationSamples)
                avg_epoch_val_loss+=loss.item()
            avg_epoch_val_loss=avg_epoch_val_loss/(i+1.0)


            if np.isnan(avg_epoch_val_loss) or np.isnan(avg_epoch_train_loss):
                print("Warning: NaN detected during inference. Model unlikely to be fully optimized!")
                break

            if monitor_validation:
                if avg_epoch_val_loss<bestModelScore:
                    bestModelState = self.basic_net.package_state()
                    bestModelScore = avg_epoch_val_loss
                    early_stop_count=0
                else:
                    early_stop_count+=1 
            else:
                if avg_epoch_train_loss<bestModelScore:
                    bestModelState = self.basic_net.package_state()
                    bestModelScore = avg_epoch_train_loss
                    early_stop_count=0
                else:
                    early_stop_count+=1 




            #track overall convergence
            trainLoss+=[avg_epoch_train_loss]
            validationLoss+=[avg_epoch_val_loss]

            error_window.append(_rel_diff(avg_epoch_train_loss,prev_train_loss))
            elbo_window.append(avg_epoch_train_loss)

            avg_error = sum(error_window)/len(error_window)
            med_error = sorted(error_window)[int(0.5*len(error_window))]
            slope_error = _returnELBOSlopeError(elbo_window)

            prev_train_loss=avg_epoch_train_loss
            errorVec+=[min([avg_error,med_error,slope_error])]





            if early_stop_count>early_stop_patience:
                if verbose:
                    print("BasicG2PModel model diverging after %03d epochs; stopping optimization. Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("BasicG2PModel model diverging after %03d epochs; stopping optimization. Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                break

            if (avg_error < errorTol) or (med_error < errorTol) or (slope_error<errorTol):


                if verbose:
                    print("BasicG2PModel model optimization converged in %03d epochs; Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("BasicG2PModel model optimization converged in %03d epochs; Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                break

            else:
                if verbose:
                    print("Completed %03d epochs of BasicG2PModel model training; Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))
                if logFile!=None:
                    with open(logFile, "a") as f:
                        f.write("Completed %03d epochs of BasicG2PModel model training; Current Loss (Train, Validation): %.4f, %.4f; Error: %.4e\n"%(epoch+1,trainLoss[-1],validationLoss[-1],errorVec[-1]))


        if epoch == (max_epochs-1):
            print('Warning: Optimization did not converge in allotted epochs')

        self.basic_net.load_state(bestModelState)
        #set eval mode
        if compute_device.split(':')[0]!='cpu':
            self.basic_net.to('cpu')
            self.cut_points=self.cut_points.detach().to('cpu')
        self.basic_net.eval()

        return bestModelScore,trainLoss,validationLoss,errorVec
            

    def PredictOutcomes(self,index=None,gene_list=None):
        assert (index is not None) ^ (gene_list is not None) or (embed_values is not None),"Must provide either: disease index, HGNC id, or a vector of embedding values to generate data."
        if self.basic_net.training:
            self.basic_net.eval()
            in_training=True
        else:
            in_training=False

        if index is not None:
            gene_list=[self.g2p_dataset.inverse_gene_map[x] for x in self.g2p_dataset.disease_table.loc[index]['MappedGenes'].values]
        assert len(set(gene_list).difference(self.g2p_dataset.gene_map.keys()))==0, "Genes: {0:s} not in GeneToPhenotypeDataset".format(','.join(list(gene_list)))
        gene_input=self.g2p_dataset.ReturnGeneDataArrays(gene_list)
        output=self.basic_net.forward(gene_input)

      
        if in_training:
            self.basic_net.train()
        symptoms=torch.sigmoid(output[0]).detach().clone().numpy()
        missing=torch.sigmoid(output[1]).detach().clone().numpy()
        frequency=torch.exp(dist.OrderedLogistic(output[2],self.cut_points).logits).detach().clone().numpy()  
        return {'BaselineSymptomRates':symptoms,"MissingRates":missing,'FrequencyRates':frequency}  

    def PerplexCompareNull(self,hpo_freq_pairs,gene_list):

        if self.basic_net.training:
            self.basic_net.eval()
            in_training=True
        else:
            in_training=False

        annot_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.n_pred),dtype=torch.float32)
        freq_array=torch.zeros((max(len(hpo_freq_pairs),len(gene_list)),self.n_pred),dtype=torch.float32)
        for t,hpo_freq_vec in enumerate(hpo_freq_pairs):
            idx_vec=[self.g2p_dataset.symptom_map[x[0]] for x in hpo_freq_vec]
            freq_vec=[self.g2p_dataset.ordinal_freq_map[x[1]] for x in hpo_freq_vec]
            annot_array[t,idx_vec]=1
            freq_array[t,idx_vec]=torch.tensor(freq_vec,dtype=torch.float32)
        data_arrays=(annot_array,freq_array,self.g2p_dataset.ReturnGeneDataArrays(gene_list))


        pred_log_odds=self.basic_net.forward(data_arrays[2])
        if in_training:
            self.basic_net.train()

        null_model_log_odds=tuple(torch.logit(self.g2p_dataset._torchWrapper(x)).unsqueeze(dim=0) for x in self.g2p_dataset.InferNullModel().values())
        output=np.zeros((pred_log_odds[0].shape[0],3))
        perplex_function_null=BasicG2PMapLoss_Categortical()
        perplex_function_basic=BasicG2PMapLoss()
        for i in range(output.shape[0]):
            output[i,0]=perplex_function_basic(pred_log_odds[0][i:i+1],pred_log_odds[1][i:i+1],pred_log_odds[2][i:i+1], data_arrays[0][i:i+1],data_arrays[1][i:i+1],self.cut_points).detach().item()
            output[i,1]=perplex_function_null(null_model_log_odds[0],null_model_log_odds[1],null_model_log_odds[2], data_arrays[0][i:i+1],data_arrays[1][i:i+1]).detach().item()
            output[i,2]=output[i,0]-output[i,1]

        return {'BasicG2PModel Perplex.':output[:,0],'Null Perplex.':output[:,1],'BasicG2PModel-Null':output[:,2]}


    def LoadModel(self,stored_model):
        """
        Loads previously fit model either from a dictionary (generated using PackageModel) or from a file path (with file constructed using PackageModel)

        Parameters
        ----------
        stored_model : either dict or str (file path)

        Returns
        -------
        None.

        """
        if not isinstance(stored_model,dict):
            assert isinstance(stored_model,str),"Expects file name if not provided with dictionary."
            with open(stored_model,'rb') as f:
                model_dict = torch.load(f,map_location='cpu')

        self.basic_net.load_state(model_dict)



    def PackageModel(self,fName=None):
        """
        Packages the current model and returns it as a python dictionary. Will optionally write this dictionary to disk using PyTorch.

        Parameters
        ----------
        fName : str, default None
            File path to save model to disk. The default is None, which means that only a model dictionary will be returned.

        Returns
        -------
        model_dict : dict
            Dictionary containing fitted model parameters in addition to general meta data.

        """
        model_dict = self.basic_net.package_state()
        if fName is not None:
            with open(fName,'wb') as f:
                torch.save(model_dict,f)
        return model_dict



